import { parse_script } from '@scrow/tapscript/script';
import { CoreWallet } from './wallet.js';
import { buffer_tx, create_prevout } from '@scrow/tapscript/tx';
import { parse_args, run_cmd } from './cmd.js';
import { cmd_config, core_config } from '../config.js';
import { clone, convert_value, convert_vout, now } from './util.js';
export class CoreClient {
    constructor(core, config) {
        const opt = core_config(config);
        this._cache = ['null', null];
        this.params = [
            `-chain=${opt.network}`,
            ...opt.params,
            ...opt.cli_params
        ];
        if (opt.rpc_user !== undefined) {
            this.params.push(`-rpcuser=${opt.rpc_user}`);
        }
        if (opt.rpc_pass !== undefined) {
            this.params.push(`-rpcpassword=${opt.rpc_pass}`);
        }
        if (opt.rpc_port !== undefined) {
            this.params.push(`-rpcport=${opt.rpc_port}`);
        }
        if (opt.confpath !== undefined) {
            this.params.push(`-conf=${opt.confpath}`);
        }
        if (opt.datapath !== undefined) {
            this.params.push(`-datadir=${opt.datapath}`);
        }
        if (opt.cookiepath !== undefined) {
            this.params.push(`-rpccookiefile=${opt.cookiepath}`);
        }
        this._opt = opt;
        this._core = core;
        this._faucet = null;
        this._debug('[client] Initializing with params:', this.params.join(' '));
    }
    get opt() {
        return this._opt;
    }
    get blocks() {
        return this.cmd('getblockcount');
    }
    get chain_info() {
        return this.cmd('getblockchaininfo');
    }
    get core() {
        return this._core;
    }
    get wallets_loaded() {
        return this.cmd('listwallets', null, { cache: true });
    }
    get wallets_created() {
        return this.cmd('listwalletdir', null, { cache: true })
            .then(e => e.wallets.map(x => x.name));
    }
    _debug(...msg) {
        if (this.opt.debug) {
            console.log('[client]', ...msg);
        }
    }
    _log(...msg) {
        if (this.opt.verbose) {
            console.log('[client]', ...msg);
        }
    }
    async cmd(method, args, config) {
        const { clipath = 'bitcoin-cli', debug, use_cache } = this.opt;
        const { cache, params } = cmd_config(config);
        const parsed = parse_args(method, args);
        const witness = [...this.params, ...params, ...parsed];
        const label = witness.join('');
        if (debug) {
            const offset = this.params.length;
            this._debug('cmd:', witness.slice(offset).join(' '));
        }
        if (cache && use_cache && this._cache[0] === label) {
            this._debug('using cache for method:', method);
            return clone(this._cache[1]);
        }
        const data = await run_cmd(clipath, witness);
        this._cache = [label, data];
        return clone(data);
    }
    async _get_block_data(query, txdata = false) {
        let { height, hash } = query;
        if (height === undefined && hash === undefined) {
            height = await this.cmd('getblockcount');
        }
        if (typeof height === 'number') {
            hash = await this.cmd('getblockhash', height);
        }
        if (typeof hash !== 'string') {
            throw new Error('Unable to fetch any blocks!');
        }
        return (txdata === true)
            ? this.cmd('getblock', hash)
            : this.cmd('getblockheader', hash);
    }
    async get_block(query) {
        return this._get_block_data(query, true);
    }
    async get_header(query) {
        return this._get_block_data(query, false);
    }
    async scan_txout(action, ...desc) {
        return this.cmd('scantxoutset', [action, JSON.stringify(desc)]);
    }
    async mine_blocks(count = 1, addr) {
        if (this.opt.network !== 'regtest') {
            throw new Error('You can only generate funds on regtest network!');
        }
        if (addr === undefined) {
            addr = await this.core.faucet.get_address('faucet');
        }
        return this.cmd('generatetoaddress', [count, addr]);
    }
    async get_tx(txid) {
        try {
            const res = await this.cmd('getrawtransaction', [txid, 2]);
            res.vout = convert_vout(res.vout);
            return res;
        }
        catch {
            return null;
        }
    }
    async get_tx_status(txid) {
        const tx = await this.get_tx(txid);
        if (tx === null)
            return null;
        const blocks = await this.blocks;
        return get_tx_status(blocks, tx);
    }
    async get_txout(txid, vout) {
        const res = await this.cmd('gettxout', [txid, vout]);
        if (res === null)
            return null;
        const value = convert_value(res.value);
        return { ...res, value };
    }
    async get_utxos(opt) {
        const desc = get_scan_desc(opt);
        return this.scan_txout('start', desc).then(res => {
            const { success, unspents } = res;
            if (!success)
                return [];
            return unspents.map(e => {
                return { ...e, amount: convert_value(e.amount) };
            });
        });
    }
    async get_txout_status(txid, vout) {
        const txout = await this.get_txout(txid, vout);
        if (txout === null)
            return null;
        const address = txout.scriptPubKey?.address;
        if (address === undefined)
            return null;
        if (txout.confirmations === 0) {
            return { spent: false };
        }
        const utxos = await this.get_utxos({ address });
        let utxo = utxos.find(e => e.txid === txid && e.vout === vout);
        if (utxo === undefined) {
            return { spent: true };
        }
        else {
            return { spent: false };
        }
    }
    async get_txinput(txid, vout) {
        const tx = await this.get_tx(txid);
        if (tx === null)
            return null;
        const txout = tx.vout.at(vout);
        if (txout === undefined)
            return null;
        const blocks = await this.blocks;
        const status = get_tx_status(blocks, tx);
        const { value, scriptPubKey } = txout;
        const script = parse_script(scriptPubKey.hex);
        const prevout = { value, scriptPubKey: script.asm };
        const txinput = create_prevout({ txid, vout, prevout });
        return { txinput, status };
    }
    async load_wallet(name) {
        const wallet = new CoreWallet(this, name);
        await wallet.load();
        await wallet.init();
        return wallet;
    }
    async load_wallets(...labels) {
        const wallets = {};
        const files = await this.wallets_created;
        const names = await this.wallets_loaded;
        for (const label of labels) {
            const wallet = new CoreWallet(this, label);
            if (!names.includes(label)) {
                if (!files.includes(label)) {
                    await wallet._create();
                }
                else {
                    await wallet._load();
                }
            }
            wallets[label] = wallet;
        }
        return wallets;
    }
    async publish_tx(txdata, confirm = false) {
        const txhex = buffer_tx(txdata).hex;
        const txid = await this.cmd('sendrawtransaction', [txhex]);
        if (confirm)
            await this.mine_blocks(1);
        return txid;
    }
    async set_time(timestamp) {
        if (this.opt.network !== 'regtest') {
            throw new Error('You can only manipulate time on regtest network!');
        }
        const ts = timestamp ?? now();
        await this.cmd('setmocktime', [ts]);
    }
}
function get_scan_desc(opt) {
    if (opt.address !== undefined) {
        return `addr(${opt.address})`;
    }
    else if (opt.pubkey !== undefined) {
        return `combo(${opt.pubkey})`;
    }
    else if (opt.script) {
        return `raw(${opt.script})`;
    }
    throw new Error('No scan option specified!');
}
function get_tx_status(blocks, tx) {
    if (tx.confirmations !== undefined && tx.confirmations > 0) {
        const block_hash = tx.blockhash;
        const block_time = tx.blocktime;
        const block_height = blocks - tx.confirmations;
        return { confirmed: true, block_height, block_hash, block_time };
    }
    else {
        return { confirmed: false };
    }
}
//# sourceMappingURL=client.js.map